"""Listing pull requests."""
import contextlib
from typing import Collection
from typing import Iterator
from typing import Optional

from retrocookie.pr import events
from retrocookie.pr.base.bus import Bus
from retrocookie.pr.protocols import github


def get_pull_request(
    repository: github.Repository,
    spec: str,
    *,
    bus: Bus,
) -> github.PullRequest:
    """Return the pull request matching the provided spec."""
    with contextlib.suppress(ValueError):
        number = int(spec)
        return repository.pull_request(number)

    pull = repository.pull_request_by_head(f"{repository.owner}:{spec}")
    if pull is not None:
        return pull

    bus.events.raise_(events.PullRequestNotFound(spec))


def get_pull_requests(
    repository: github.Repository,
    specs: Collection[str] = (),
    *,
    bus: Bus,
) -> Iterator[github.PullRequest]:
    """Return pull requests. With specs, filter those matching specs."""
    if specs:
        for spec in specs:
            yield get_pull_request(repository, spec, bus=bus)
    else:
        yield from repository.pull_requests()


def list_pull_requests(
    repository: github.Repository,
    specs: Collection[str] = (),
    *,
    user: Optional[str] = None,
    bus: Bus,
) -> Iterator[github.PullRequest]:
    """List matching pull requests in repository."""
    for pull in get_pull_requests(repository, specs, bus=bus):
        if user is None or pull.user == user:
            yield pull



"""Compatibility shim for contextlib."""
import contextlib
from typing import Callable
from typing import ContextManager
from typing import Iterator
from typing import TypeVar


T = TypeVar("T")


def contextmanager(
    func: Callable[..., Iterator[T]]
) -> Callable[..., ContextManager[T]]:
    """Fix annotations of functions decorated by contextlib.contextmanager."""
    result = contextlib.contextmanager(func)
    result.__annotations__ = {**func.__annotations__, "return": ContextManager[T]}
    return result



"""Core module."""
import json
import tempfile
from pathlib import Path
from typing import Any
from typing import cast
from typing import Container
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional

from . import git
from .filter import RepositoryFilter
from retrocookie.compat import contextlib


@contextlib.contextmanager
def temporary_repository() -> Iterator[git.Repository]:
    """Create repository in temporary directory."""
    with tempfile.TemporaryDirectory() as tmpdir:
        path = Path(tmpdir)
        yield git.Repository.init(path)


def find_template_directory(repository: git.Repository) -> Path:
    """Locate the subdirectory with the project template."""
    tokens = "{{", "cookiecutter", "}}"
    for path in repository.path.iterdir():
        if path.is_dir() and all(x in path.name for x in tokens):
            return path.relative_to(repository.path)
    raise Exception("cannot find template directory")


def load_context(repository: git.Repository, ref: str) -> Dict[str, Any]:
    """Load the context from the .cookiecutter.json file."""
    path = Path(".cookiecutter.json")
    text = repository.read_text(path, ref=ref)
    data = json.loads(text)
    if not isinstance(data, dict) or not all(isinstance(key, str) for key in data):
        raise TypeError(".cookiecutter.json does not contain an object")
    return cast(Dict[str, Any], data)


def get_commits(
    repository: git.Repository,
    commits: Iterable[str],
    branch: Optional[str],
    upstream: str,
) -> List[str]:
    """Return hashes of the commits to be picked."""

    def _generate() -> Iterator[str]:
        if branch is not None:
            yield from repository.parse_revisions(f"{upstream}..{branch}")

        if commits:
            yield from repository.parse_revisions(*commits)

    return list(_generate())


def rewrite_commits(
    repository: git.Repository,
    source: git.Repository,
    template_directory: Path,
    include_variables: Container[str],
    exclude_variables: Container[str],
    commits: Iterable[str],
) -> List[str]:
    """Rewrite the repository using template variables."""
    commits = list(commits)
    with_parents = source.parse_revisions(
        *[f"{commit}^" for commit in commits], *commits
    )
    context = load_context(source, commits[-1])
    RepositoryFilter(
        repository=repository,
        source=source,
        commits=with_parents,
        template_directory=template_directory,
        context=context,
        include_variables=include_variables,
        exclude_variables=exclude_variables,
    ).run()
    return [repository.lookup_replacement(commit) for commit in commits]


def apply_commits(
    repository: git.Repository,
    source: git.Repository,
    commits: Iterable[str],
    create_branch: Optional[str],
) -> None:
    """Create <branch> with the specified commits from <remote>."""
    repository.fetch_commits(source, *commits)

    if create_branch:
        repository.create_branch(create_branch)
        repository.switch_branch(create_branch)

    repository.cherrypick(*commits)


def retrocookie(
    instance_path: Path,
    commits: Iterable[str] = (),
    *,
    path: Optional[Path] = None,
    branch: Optional[str] = None,
    upstream: Optional[str] = None,
    create_branch: Optional[str] = None,
    include_variables: Container[str] = (),
    exclude_variables: Container[str] = (),
) -> None:  # noqa: DAR101 https://github.com/terrencepreilly/darglint/issues/56
    """Import commits from instance repository into template repository.

    This function imports commits from an instance of the Cookiecutter
    template, rewriting occurrences of template variables back into the
    original templating tags, and prepending the template directory to
    filenames. Any tokens with special meaning in Jinja are escaped.

    Args:
        instance_path: The source repository, an instance of the Cookiecutter
            template with a ``.cookiecutter.json`` file.

        commits: The commits to be imported from the source repository.

        path: The target repository, a Cookiecutter template. This defaults to
            the current working directory.

        branch: The name of a branch to be imported from the source repository.
            This is equivalent to passing ``["master..branch"]`` in the
            ``commits`` parameter.

        upstream: The upstream for ``branch``. This defaults to the value of the
            ``init.defaultBranch`` option in git. If the option is not set,
            ``master`` is used.

        create_branch: The name of the branch to be created in the target
            repository. By default, commits are imported onto the current
            branch.

        include_variables: The Cookiecutter variables which should be
            rewritten. If this is not specified, all variables from
            cookiecutter.json are rewritten.

        exclude_variables: Any Cookiecutter variables which should not be
            rewritten.

    """
    if upstream is None:
        upstream = git.get_default_branch()

    repository = git.Repository(path)
    instance = git.Repository(instance_path)
    template_directory = find_template_directory(repository)
    commits = get_commits(instance, commits, branch, upstream)

    with temporary_repository() as scratch:
        commits = rewrite_commits(
            scratch,
            instance,
            template_directory,
            include_variables,
            exclude_variables,
            commits,
        )

        apply_commits(repository, scratch, commits, create_branch)



"""Application cache."""
import hashlib
import json
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Iterator

from retrocookie import git
from retrocookie.compat import contextlib


# Windows does not properly support files and directories longer than
# 260 characters. Use a smaller digest size on this platform.
DIGEST_SIZE: int = 64 if sys.platform != "win32" else 32


@dataclass
class Cache:
    """Application cache."""

    path: Path

    def save_token(self, token: str) -> None:
        """Save a token."""
        data = {"token": token}
        text = json.dumps(data)
        path = self.path / "token.json"
        path.parent.mkdir(exist_ok=True, parents=True)
        path.touch(mode=0o600)
        path.write_text(text)

    def load_token(self) -> str:
        """Load a token."""
        path = self.path / "token.json"
        text = path.read_text()
        data = json.loads(text)
        token = data["token"]
        if not isinstance(token, str):
            raise TypeError("invalid token")
        return token

    def _repository_path(self, url: str) -> Path:
        h = hashlib.blake2b(url.encode(), digest_size=DIGEST_SIZE).hexdigest()
        return self.path / "repositories" / h[:2] / h[2:] / "repo.git"

    def repository(self, url: str) -> git.Repository:
        """Clone or update repository."""
        path = self._repository_path(url)

        if not path.exists():
            return git.Repository.clone(url, path, mirror=True)

        repository = git.Repository(path)
        repository.update_remote()
        return repository

    @contextlib.contextmanager
    def worktree(
        self,
        repository: git.Repository,
        branch: str,
        *,
        base: str = "HEAD",
        force: bool = False,
    ) -> Iterator[git.Repository]:
        """Context manager to add and remove a worktree."""
        assert self.path in repository.path.parents  # noqa: S101
        path = repository.path.parent / "worktrees" / branch

        with repository.worktree(
            branch, path, base=base, force=force, force_remove=True
        ) as worktree:
            yield worktree



"""Import pull requests."""
from typing import Collection
from typing import Optional
from urllib.parse import urlparse

from retrocookie import git
from retrocookie.core import load_context
from retrocookie.pr import events
from retrocookie.pr.base.bus import Bus
from retrocookie.pr.base.exceptionhandlers import ExceptionHandler
from retrocookie.pr.cache import Cache
from retrocookie.pr.importer import Importer
from retrocookie.pr.list import list_pull_requests
from retrocookie.pr.protocols import github
from retrocookie.pr.protocols.retrocookie import Retrocookie
from retrocookie.pr.repository import Repository
from retrocookie.utils import removeprefix
from retrocookie.utils import removesuffix


def parse_repository_name(url: str) -> Optional[str]:
    """Extract the repository name from the URL, if possible."""
    for prefix in ("gh:", "git@github.com:"):
        if url.startswith(prefix):
            path = removeprefix(url, prefix)
            return removesuffix(path, ".git")

    result = urlparse(url)
    if result.hostname == "github.com":
        return removesuffix(result.path[1:], ".git")

    return None


def get_project_name(*, bus: Bus) -> str:
    """Return the repository name of the project."""
    with bus.events.reraise(events.ProjectNotFound()):
        repository = git.Repository()

    for remote in repository.repo.remotes:
        name = parse_repository_name(remote.url)
        if name is not None:
            return name

    bus.events.raise_(events.ProjectNotFound())


def get_template_name(project: Repository, *, bus: Bus) -> str:
    """Return the repository name of the template, given the project."""
    with bus.events.reraise(
        events.TemplateNotFound(project.github),
        when=(KeyError, TypeError),
    ):
        context = load_context(project.clone, "HEAD")

    name = parse_repository_name(context["_template"])
    if name is not None:
        return name

    bus.events.raise_(events.TemplateNotFound(project.github))


def check_git_version(*, bus: Bus) -> None:
    """Check if we have a recent version of git."""
    # newren/git-filter-repo requires git >= 2.22.0
    minimum_version = git.Version.parse("2.22.0")

    with bus.events.reraise(events.GitNotFound()):
        version = git.version()

    if version < minimum_version:
        bus.events.raise_(events.BadGitVersion(version, minimum_version))


def import_pull_requests(
    pull_requests: Collection[str] = (),
    *,
    api: github.API,
    bus: Bus,
    cache: Cache,
    errorhandler: ExceptionHandler,
    retrocookie: Retrocookie,
    repository: Optional[str] = None,
    user: Optional[str] = None,
    force: bool = False,
) -> None:
    """Import pull requests from a repository into its Cookiecutter template."""
    check_git_version(bus=bus)

    if repository is None:
        repository = get_project_name(bus=bus)
    elif "/" not in repository:
        repository = "/".join([api.me, repository])

    with bus.contexts.publish(events.LoadProject(repository)):
        project = Repository.load(repository, api=api, cache=cache)

    template_name = get_template_name(project, bus=bus)

    with bus.contexts.publish(events.LoadTemplate(template_name)):
        template = Repository.load(template_name, api=api, cache=cache)

    importer = Importer(
        project,
        template,
        bus=bus,
        cache=cache,
        retrocookie=retrocookie,
    )

    for pull in list_pull_requests(project.github, pull_requests, user=user, bus=bus):
        with errorhandler:
            importer.import_(pull, force=force)



"""Application console."""
from typing import ContextManager
from typing import Iterator

import rich.console
import rich.theme
import rich.traceback
from rich.markup import escape

from retrocookie.compat import contextlib
from retrocookie.pr import events
from retrocookie.pr.base.bus import Bus
from retrocookie.pr.compat import shlex


class Console:
    """Console."""

    def __init__(self) -> None:
        """Create the console."""
        theme = rich.theme.Theme(
            {
                "success": "green",
                "failure": "red",
                "hint": "yellow",
                "repository": "blue",
                "pull": "bold bright_white",
                "oldpull": "bright_black",
                "title": "cyan",
                "command": "bold",
                "status": "bold",
                "stdout": "bright_black",
                "stderr": "red",
                "version": "green",
                "http.method": "blue",
                "http.status": "bright_blue",
            }
        )
        self.console = rich.console.Console(stderr=True, highlight=False, theme=theme)

    def success(self, message: str) -> None:
        """Display a success message."""
        self.console.print(f"[success]✓[/] {message}")

    def failure(self, message: str) -> None:
        """Display a failure message."""
        self.console.print(f"[failure]⨯[/] {message}")

    def hint(self, message: str) -> None:
        """Display a hint for the user."""
        self.console.print(f"[hint]☞ {message}[/]")

    def highlight(self, text: str) -> None:
        """Output a text with highlighting."""
        console = rich.console.Console(stderr=True)
        console.print(text)

    @contextlib.contextmanager
    def progress(self, message: str) -> Iterator[None]:
        """Display a status message with a spinner."""
        try:
            with self.console.status(message):
                yield
        except Exception:
            self.failure(message)
            raise


def start(*, bus: Bus) -> None:
    """Create the console and subscribe to events."""
    rich.traceback.install()

    console = Console()

    _subscribe(console, bus)


def _subscribe(console: Console, bus: Bus) -> None:  # noqa: C901
    _ = None  # Ignoring redefinitions of this name below.

    @bus.events.subscribe  # type: ignore[no-redef]
    def _(event: events.GitNotFound) -> None:
        console.failure("git not found")
        console.hint("Do you have git installed?")
        console.hint("Is git on your PATH?")
        console.hint("Can you run [command]git version[/]?")

    @bus.events.subscribe  # type: ignore[no-redef]
    def _(event: events.BadGitVersion) -> None:
        console.failure(f"This program requires git [version]{event.expected}[/].")
        console.hint(f"You have git [version]{event.version}[/], which is too old.")

    @bus.events.subscribe  # type: ignore[no-redef]
    def _(event: events.GitFailed) -> None:
        def _lines() -> Iterator[str]:
            yield (
                f"The command [command]git {event.command}[/] failed"
                f" with status [status]{event.status}[/]."
            )

            command = escape(shlex.join(["git", event.command, *event.options]))

            yield ""
            yield f"❯ {command}"

            for stream in ["stdout", "stderr"]:
                output = getattr(event, stream)
                if output:
                    yield ""
                    for line in output.splitlines():
                        line = escape(line)
                        yield f"[{stream}]{line}[/]"

            yield ""

        console.failure("\n".join(_lines()))

        if event.command == "cherry-pick":
            console.hint("Looks like the changes did not apply cleanly.")
            console.hint(
                "Try running [command]retrocookie[/] on a local clone instead."
            )

    @bus.events.subscribe  # type: ignore[no-redef]
    def _(event: events.GitHubError) -> None:
        def _lines() -> Iterator[str]:
            yield "The GitHub API returned an error response."
            yield f"[http.method]{event.method}[/] [repr.url]{event.url}[/]"
            yield f"[http.status]{event.code}[/] {event.message}"
            yield from event.errors

        console.failure("\n".join(_lines()))

    @bus.events.subscribe  # type: ignore[no-redef]
    def _(event: events.ConnectionError) -> None:
        def _lines() -> Iterator[str]:
            yield "Connection to GitHub API could not be established."
            yield f"[http.method]{event.method}[/] [repr.url]{event.url}[/]"

        console.failure("\n".join(_lines()))
        console.highlight(event.error)

    @bus.events.subscribe  # type: ignore[no-redef]
    def _(event: events.ProjectNotFound) -> None:
        console.failure("Project not found")
        console.hint("Does the current directory contain a repository?")
        console.hint("Is the repository on GitHub?")

    @bus.events.subscribe  # type: ignore[no-redef]
    def _(event: events.TemplateNotFound) -> None:
        console.failure(
            f"Project template for [repository]{event.project.full_name}[/] not found"
        )
        console.hint("Does the project contain a .cookiecutter.json file?")
        console.hint(
            'Does the .cookiecutter.json contain the repository URL under "_template"?'
        )
        console.hint("Is the template repository on GitHub?")

    @bus.contexts.subscribe  # type: ignore[no-redef]
    def _(event: events.LoadProject) -> ContextManager[None]:
        return console.progress(f"Loading project [repository]{event.repository}[/]")

    @bus.contexts.subscribe  # type: ignore[no-redef]
    def _(event: events.LoadTemplate) -> ContextManager[None]:
        return console.progress(f"Loading template [repository]{event.repository}[/]")

    @bus.events.subscribe  # type: ignore[no-redef]
    def _(event: events.RepositoryNotFound) -> None:
        console.failure(f"Repository [repository]{event.repository}[/] not found")

    @bus.events.subscribe  # type: ignore[no-redef]
    def _(event: events.PullRequestNotFound) -> None:
        spec = event.pull_request
        if spec.isnumeric():
            spec = f"#{spec}"
        console.failure(f"Pull request [pull]{spec}[/] not found")

    @bus.events.subscribe  # type: ignore[no-redef]
    def _(event: events.PullRequestAlreadyExists) -> None:
        console.failure(
            "Already imported"
            f" [pull]#{event.project_pull.number}[/]"
            f" [title]{escape(event.project_pull.title)}[/]"
        )
        console.hint(f"See [pull]#{event.template_pull.number}[/]")
        console.hint("Use --force to update an existing pull request.")

    @bus.contexts.subscribe  # type: ignore[no-redef]
    def _(event: events.CreatePullRequest) -> ContextManager[None]:
        return console.progress(
            f"[pull]#{event.project_pull.number}[/]"
            f" [title]{escape(event.project_pull.title)}[/]"
        )

    @bus.contexts.subscribe  # type: ignore[no-redef]
    @contextlib.contextmanager
    def _(event: events.UpdatePullRequest) -> Iterator[None]:
        with console.progress(
            f"[pull]#{event.project_pull.number}[/]"
            f" [title]{escape(event.project_pull.title)}[/]"
            f" [oldpull]#{event.template_pull.number}[/]"
        ):
            yield

        console.success(
            f"[pull]#{event.project_pull.number}[/]"
            f" [title]{escape(event.project_pull.title)}[/]"
            f" [pull]#{event.template_pull.number}[/]"
        )

    @bus.events.subscribe  # type: ignore[no-redef]
    def _(event: events.PullRequestCreated) -> None:
        console.success(
            f"[pull]#{event.project_pull.number}[/]"
            f" [title]{escape(event.project_pull.title)}[/]"
            f" [pull]#{event.template_pull.number}[/]"
        )



"""GitHub interface."""
from __future__ import annotations

from typing import AbstractSet
from typing import cast
from typing import Iterable
from typing import Optional
from typing import Set
from typing import Union

import github3.exceptions
import requests
import tenacity

from retrocookie.pr import events
from retrocookie.pr.base.bus import Bus
from retrocookie.pr.base.exceptionhandlers import ExceptionHandler
from retrocookie.pr.base.exceptionhandlers import exceptionhandler


class PullRequest:
    """Pull request in a GitHub repository."""

    def __init__(
        self,
        pull_request: Union[
            github3.pulls.PullRequest,
            github3.pulls.ShortPullRequest,
        ],
    ) -> None:
        """Initialize."""
        self._pull_request = pull_request

    @property
    def number(self) -> int:
        """Number of the pull request."""
        return cast(int, self._pull_request.number)

    @property
    def title(self) -> str:
        """Title of the pull request description."""
        return cast(str, self._pull_request.title)

    @property
    def body(self) -> Optional[str]:
        """Body of the pull request description."""
        return cast(Optional[str], self._pull_request.body)

    @property
    def branch(self) -> str:
        """Branch merged by the pull request."""
        return cast(str, self._pull_request.head.ref)

    @property
    def user(self) -> str:
        """Login of the user that opened the pull request."""
        return cast(str, self._pull_request.user.login)

    @property
    def html_url(self) -> str:
        """URL for viewing the pull request in a browser."""
        return cast(str, self._pull_request.html_url)

    @property
    def labels(self) -> Set[str]:
        """The labels associated with the pull request."""
        issue = self._pull_request.issue()
        return {label.name for label in issue.labels()}

    def update(self, title: str, body: Optional[str], labels: AbstractSet[str]) -> None:
        """Update the pull request."""
        self._pull_request.update(title=title, body=body)
        self._pull_request.issue().replace_labels([*labels])


class Repository:
    """GitHub Repository."""

    def __init__(
        self, repository: github3.repos.repo.Repository, *, token: str, bus: Bus
    ) -> None:
        """Initialize."""
        self._repository = repository
        self._token = token
        self._bus = bus

    @property
    def owner(self) -> str:
        """The user who owns the repository."""
        return cast(str, self._repository.owner.login)

    @property
    def full_name(self) -> str:
        """The full name of the repository."""
        return cast(str, self._repository.full_name)

    @property
    def clone_url(self) -> str:
        """URL for cloning the repository."""
        return cast(str, self._repository.clone_url)

    @property
    def push_url(self) -> str:
        """URL for pushing to the repository."""
        return f"https://{self._token}:x-oauth-basic@github.com/{self.full_name}.git"

    @property
    def default_branch(self) -> str:
        """The default branch of the repository."""
        return cast(str, self._repository.default_branch)

    def pull_request(self, number: int) -> PullRequest:
        """Return pull request identified by the given number."""
        with self._bus.events.reraise(
            events.PullRequestNotFound(str(number)),
            when=github3.exceptions.NotFoundError,
        ):
            pull_request = self._repository.pull_request(number)

        return PullRequest(pull_request)

    def pull_request_by_head(self, head: str) -> Optional[PullRequest]:
        """Return pull request for the given head."""
        for pull_request in self._repository.pull_requests(head=head):
            return PullRequest(pull_request)
        return None

    def pull_requests(self) -> Iterable[PullRequest]:
        """Pull requests open in the repository."""
        for pull_request in self._repository.pull_requests(state="open"):
            yield PullRequest(pull_request)

    def create_pull_request(
        self,
        *,
        head: str,
        title: str,
        body: Optional[str],
        labels: AbstractSet[str],
    ) -> PullRequest:
        """Create a pull request."""
        pull_request = self._repository.create_pull(
            title=title,
            body=body,
            head=head,
            base=self._repository.default_branch,
        )

        # The GitHub API sometimes responds with 404 when the issue for
        # a newly created pull request is requested.
        for attempt in tenacity.Retrying(
            reraise=True,
            stop=tenacity.stop_after_attempt(3),  # type: ignore[attr-defined]
            wait=tenacity.wait_fixed(3),  # type: ignore[attr-defined]
        ):
            with attempt:
                pull_request.issue().add_labels(*labels)

        return PullRequest(pull_request)


class API:
    """GitHub API."""

    def __init__(self, github: github3.GitHub, *, token: str, bus: Bus) -> None:
        """Initialize."""
        self._github = github
        self._token = token
        self._bus = bus

    @classmethod
    def login(cls, token: str, *, bus: Bus) -> API:
        """Login to GitHub."""
        github = github3.login(token=token)
        return cls(github, token=token, bus=bus)

    @property
    def me(self) -> str:
        """Return the login of the authenticated user."""
        return cast(str, self._github.me().login)

    def repository(self, owner: str, name: str) -> Repository:
        """Return the repository with the given owner and name."""
        fullname = "/".join([owner, name])
        with self._bus.events.reraise(
            events.RepositoryNotFound(fullname),
            when=github3.exceptions.NotFoundError,
        ):
            _repository = self._github.repository(owner, name)

        return Repository(_repository, token=self._token, bus=self._bus)


def errorhandler(*, bus: Bus) -> ExceptionHandler:
    """Handle GitHub error responses."""

    @exceptionhandler
    def _handler1(error: github3.exceptions.GitHubError) -> None:
        if error.response is not None and error.response.request is not None:
            bus.events.raise_(
                events.GitHubError(
                    error.response.request.url,
                    error.response.request.method,
                    error.code,
                    error.message,  # noqa: B306
                    error.errors,
                )
            )
        return None

    @exceptionhandler
    def _handler2(error: github3.exceptions.ConnectionError) -> None:
        for arg in error.args:
            if isinstance(arg, requests.RequestException):
                bus.events.raise_(
                    events.ConnectionError(
                        arg.request.url,
                        arg.request.method,
                        str(error),
                    )
                )
        return None

    return _handler1 >> _handler2



"""High-level repository abstraction."""
from __future__ import annotations

from dataclasses import dataclass

from retrocookie import git
from retrocookie.pr.cache import Cache
from retrocookie.pr.protocols import github as gh


@dataclass
class Repository:
    """High-level repository abstraction."""

    github: gh.Repository
    clone: git.Repository

    @classmethod
    def load(
        cls,
        repository: str,
        *,
        api: gh.API,
        cache: Cache,
    ) -> Repository:
        """Load all the data associated with a GitHub repository."""
        owner, name = repository.split("/", 1)
        github = api.repository(owner, name)
        clone = cache.repository(github.clone_url)
        return cls(github, clone)



"""Import pull requests."""
from dataclasses import dataclass

from retrocookie.pr import appname
from retrocookie.pr import events
from retrocookie.pr.base.bus import Bus
from retrocookie.pr.base.bus import Context
from retrocookie.pr.cache import Cache
from retrocookie.pr.protocols import github
from retrocookie.pr.protocols.retrocookie import Retrocookie
from retrocookie.pr.repository import Repository


@dataclass
class Importer:
    """Importer for pull requests."""

    project: Repository
    template: Repository
    bus: Bus
    cache: Cache
    retrocookie: Retrocookie

    def import_(
        self,
        project_pull: github.PullRequest,
        *,
        force: bool = False,
    ) -> None:
        """Import the given pull request from the project into its template."""
        template_branch = f"{appname}/{project_pull.branch}"
        head = f"{self.template.github.owner}:{template_branch}"
        template_pull = self.template.github.pull_request_by_head(head)
        event: Context

        if template_pull is None:
            event = events.CreatePullRequest(self.template.github, head, project_pull)
        elif force:
            event = events.UpdatePullRequest(
                self.template.github, template_pull, project_pull
            )
        else:
            self.bus.events.raise_(
                events.PullRequestAlreadyExists(
                    self.template.github, template_pull, project_pull
                )
            )

        with self.bus.contexts.publish(event):
            with self.cache.worktree(
                self.template.clone, template_branch, force=True
            ) as worktree:
                self.retrocookie(
                    self.project.clone.path,
                    branch=project_pull.branch,
                    upstream=self.project.github.default_branch,
                    path=worktree.path,
                )
            # https://stackoverflow.com/a/41153073/1355754
            self.template.clone.push(
                self.template.github.push_url, f"+{template_branch}", force=force
            )

            if template_pull is None:
                template_pull = self.template.github.create_pull_request(
                    head=head,
                    title=project_pull.title,
                    body=project_pull.body,
                    labels=project_pull.labels,
                )
            else:
                template_pull.update(
                    title=project_pull.title,
                    body=project_pull.body,
                    labels=project_pull.labels,
                )

        if isinstance(event, events.CreatePullRequest):
            self.bus.events.publish(
                events.PullRequestCreated(
                    self.template.github, template_pull, project_pull
                )
            )



"""Protocol for retrocookie.retrocookie."""
from pathlib import Path

from retrocookie.pr.compat.typing import Protocol


class Retrocookie(Protocol):
    """The retrocookie function."""

    def __call__(
        self, repository: Path, *, branch: str, upstream: str, path: Path
    ) -> None:
        """Import commits from instance repository into template repository."""



"""Event bus.

This module defines a :class:`Bus` class for publishing events and subscribing
handlers to them.

Events must derive from :class:`Event`. Handlers are functions that take the
event as an argument. The function must have type annotations, to allow the bus
to determine which handlers to invoke when an event is published.

>>> class Timber(Event):
...     pass
...
>>> def shout(event: Timber):
...     print("Timber!")
...
>>> bus = Bus()
>>> bus.events.subscribe(shout)
>>> bus.events.publish(Timber())
Timber!

Events can be raised on the bus. A raised event is wrapped in an :class:Error
exception, causing the stack to unwind until an error handler is encounted.

>>> try:
...     with bus.events.errorhandler():
...         bus.events.raise_(Timber())
... except Error:
...     pass
Timber!

The error handler places the event on the bus, but it does not suppress the
exception.

Contexts are a special kind of event that has a duration. Handlers for contexts
must return a context manager, which is invoked on entering and exiting the
context.

>>> class Copy(Context):
...     pass
...
>>> import contextlib
...
>>> @contextlib.contextmanager
... def handle_copy(context: Copy):
...     print("starting copy...")
...     try:
...         yield
...     except Exception:
...         print("copy failed!")
...         raise
...     else:
...         print("copy complete.")
...
>>> bus.contexts.subscribe(handle_copy)
>>> with bus.contexts.publish(Copy()):
...     print("copy: file a")
...     print("copy: file b")
starting copy...
copy: file a
copy: file b
copy complete.
"""
from collections import defaultdict
from contextlib import ExitStack
from typing import Any
from typing import ContextManager
from typing import Dict
from typing import get_type_hints
from typing import Iterator
from typing import List
from typing import NoReturn
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import Union

from retrocookie.compat.contextlib import contextmanager
from retrocookie.pr.base.exceptionhandlers import ExceptionHandler
from retrocookie.pr.base.exceptionhandlers import exceptionhandler
from retrocookie.pr.compat.typing import Protocol


__all__ = [
    "Bus",
    "Context",
    "Error",
    "Event",
]


class Event:
    """An event can be published on the bus, and subscribed to."""


class Context:
    """A context is an event with a duration."""


class Error(Exception):
    """An exception for transporting an event."""

    def __init__(self, event: Event) -> None:
        """Initialize."""
        self.event = event


E = TypeVar("E", bound=Event, contravariant=True)
C = TypeVar("C", bound=Context, contravariant=True)


class _EventHandler(Protocol[E]):
    """Handler for an event."""

    def __call__(self, event: E) -> None:
        """Invoke the handler."""


class _ContextHandler(Protocol[C]):
    """Handler for a context."""

    def __call__(self, context: C) -> ContextManager[None]:
        """Invoke the handler."""


class _Events:
    """Publish and subscribe to events."""

    def __init__(self) -> None:
        """Initialize."""
        self.handlers: Dict[
            Type[Event],
            List[_EventHandler[Any]],
        ] = defaultdict(list)

    def publish(self, event: Event) -> None:
        """Publish an event on the bus."""
        handlers = self.handlers[type(event)]
        for handler in handlers:
            handler(event)

    def subscribe(self, handler: _EventHandler[E]) -> _EventHandler[E]:
        """Subscribe to an event."""
        event_type = next(hint for hint in get_type_hints(handler).values())
        self.handlers[event_type].append(handler)
        return handler

    def raise_(self, event: Event) -> NoReturn:
        """Raise an event."""
        raise Error(event)

    def reraise(
        self,
        event: Event,
        *,
        when: Union[Type[BaseException], Tuple[Type[BaseException], ...]] = (),
    ) -> ExceptionHandler:
        """Raise an event when an exception is caught."""
        if not isinstance(when, tuple):
            when = (when,)
        elif not when:
            when = (Exception,)

        @exceptionhandler(*when)
        def _(exception: BaseException) -> NoReturn:
            raise Error(event)

        return _

    def errorhandler(self) -> ExceptionHandler:
        """Publish error events on the bus."""

        @exceptionhandler
        def _(error: Error) -> None:
            self.publish(error.event)

        return _


class _Contexts:
    """Publish and subscribe to contexts."""

    def __init__(self) -> None:
        """Initialize."""
        self.handlers: Dict[
            Type[Context],
            List[_ContextHandler[Any]],
        ] = defaultdict(list)

    @contextmanager
    def publish(self, event: Context) -> Iterator[None]:
        """Publish a context."""
        handlers = self.handlers[type(event)]
        with ExitStack() as stack:
            for handler in handlers:
                stack.enter_context(handler(event))
            yield

    def subscribe(self, handler: _ContextHandler[C]) -> _ContextHandler[C]:
        """Subscribe to a context."""
        event_type = next(hint for hint in get_type_hints(handler).values())
        self.handlers[event_type].append(handler)
        return handler


class Bus:
    """Event bus."""

    def __init__(self) -> None:
        """Initialize."""
        self.events = _Events()
        self.contexts = _Contexts()



"""Utilities."""
import os
from pathlib import Path
from typing import Iterator

from retrocookie.compat import contextlib


def removeprefix(string: str, prefix: str) -> str:
    """Remove prefix from string, if present."""
    return string[len(prefix) :] if string.startswith(prefix) else string


def removesuffix(string: str, suffix: str) -> str:
    """Remove suffix from string, if present."""
    return string[: -len(suffix)] if suffix and string.endswith(suffix) else string


@contextlib.contextmanager
def chdir(path: Path) -> Iterator[None]:
    """Context manager for changing the directory."""
    cwd = Path.cwd()
    os.chdir(path)

    try:
        yield
    finally:
        os.chdir(cwd)



"""Git interface."""
from __future__ import annotations

import contextlib
import functools
import operator
import re
import subprocess  # noqa: S404
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Any
from typing import cast
from typing import Iterator
from typing import List
from typing import Optional

import pygit2

from retrocookie.compat.contextlib import contextmanager
from retrocookie.utils import removeprefix


def git(
    *args: str, check: bool = True, **kwargs: Any
) -> subprocess.CompletedProcess[str]:
    """Invoke git."""
    return subprocess.run(  # noqa: S603,S607
        ["git", *args], check=check, text=True, capture_output=True, **kwargs
    )


VERSION_PATTERN = re.compile(
    r"""
    (?P<major>\d+)\.
    (?P<minor>\d+)
    (\.(?P<patch>\d+))?
    """,
    re.VERBOSE,
)


@dataclass(frozen=True, order=True)
class Version:
    """Simplistic representation of git versions."""

    major: int
    minor: int
    patch: int
    _text: Optional[str] = field(default=None, compare=False)

    @classmethod
    def parse(cls, text: str) -> Version:
        """Extract major.minor[.patch] from the start of the text."""
        match = VERSION_PATTERN.match(text)

        if match is None:
            raise ValueError(f"invalid version {text!r}")

        parts = match.groupdict(default="0")

        return cls(
            int(parts["major"]), int(parts["minor"]), int(parts["patch"]), _text=text
        )

    def __str__(self) -> str:
        """Return the original representation."""
        return (
            self._text
            if self._text is not None
            else f"{self.major}.{self.minor}.{self.patch}"
        )


def version() -> Version:
    """Return the git version."""
    text = git("version").stdout.strip()
    text = removeprefix(text, "git version ")
    return Version.parse(text)


def get_default_branch() -> str:
    """Return the default branch for new repositories."""
    get_configs = [
        pygit2.Config.get_global_config,
        pygit2.Config.get_system_config,
    ]
    for get_config in get_configs:
        with contextlib.suppress(IOError, KeyError):
            config = get_config()
            branch = config["init.defaultBranch"]
            assert isinstance(branch, str)  # noqa: S101
            return branch

    return "master"


class Repository:
    """Git repository."""

    def __init__(
        self, path: Optional[Path] = None, *, repo: Optional[pygit2.Repository] = None
    ) -> None:
        """Initialize."""
        if repo is None:
            self.path = path or Path.cwd()
            self.repo = pygit2.Repository(self.path)
        else:
            self.path = Path(repo.workdir or repo.path)
            self.repo = repo

    def git(self, *args: str, **kwargs: Any) -> subprocess.CompletedProcess[str]:
        """Invoke git."""
        return git(*args, cwd=self.path, **kwargs)

    @classmethod
    def init(cls, path: Path, *, bare: bool = False) -> Repository:
        """Create a repository."""
        # https://github.com/libgit2/libgit2/issues/2849
        path.parent.mkdir(exist_ok=True, parents=True)
        repo = pygit2.init_repository(path, bare=bare)
        return cls(path, repo=repo)

    @classmethod
    def clone(cls, url: str, path: Path, *, mirror: bool = False) -> Repository:
        """Clone a repository."""
        options = ["--mirror"] if mirror else []
        git("clone", *options, url, str(path))
        return cls(path)

    def create_branch(self, branch: str, ref: str = "HEAD") -> None:
        """Create a branch."""
        commit = self.repo.revparse_single(ref)
        self.repo.branches.create(branch, commit)

    def get_current_branch(self) -> str:
        """Return the current branch."""
        return self.repo.head.shorthand  # type: ignore[no-any-return]

    def exists_branch(self, branch: str) -> bool:
        """Return True if the branch exists."""
        return branch in self.repo.branches

    def switch_branch(self, branch: str) -> None:
        """Switch the current branch."""
        self.repo.checkout(self.repo.branches[branch])

    def update_remote(self) -> None:
        """Update the remotes."""
        self.git("remote", "update")

    def fetch_commits(self, source: Repository, *commits: str) -> None:
        """Fetch the given commits and their immediate parents."""
        path = source.path.resolve()
        self.git("fetch", "--no-tags", "--depth=2", str(path), *commits)

    def push(self, remote: str, *refs: str, force: bool = False) -> None:
        """Update remote refs."""
        options = ["--force-with-lease"] if force else []
        self.git("push", *options, remote, *refs)

    def parse_revisions(self, *revisions: str) -> List[str]:
        """Parse revisions using the format specified in gitrevisions(7)."""
        process = self.git("rev-list", "--no-walk", *revisions)
        result = process.stdout.split()
        result.reverse()
        return result

    def lookup_replacement(self, commit: str) -> str:
        """Lookup the replace ref for the given commit."""
        refname = f"refs/replace/{commit}"
        ref = self.repo.lookup_reference(refname)
        return cast(str, ref.target.hex)

    def _ensure_relative(self, path: Path) -> Path:
        """Interpret the path relative to the repository root."""
        return path.relative_to(self.path) if path.is_absolute() else path

    def read_text(self, path: Path, *, ref: str = "HEAD") -> str:
        """Return the contents of the blob at the given path."""
        commit = self.repo.revparse_single(ref)
        path = self._ensure_relative(path)
        blob = functools.reduce(operator.truediv, path.parts, commit.tree)
        return cast(str, blob.data.decode())

    def exists(self, path: Path, *, ref: str = "HEAD") -> bool:
        """Return True if a blob exists at the given path."""
        commit = self.repo.revparse_single(ref)
        path = self._ensure_relative(path)
        try:
            functools.reduce(operator.truediv, path.parts, commit.tree)
            return True
        except KeyError:
            return False

    def add(self, *paths: Path) -> None:
        """Add paths to the index."""
        for path in paths:
            path = self._ensure_relative(path)
            self.repo.index.add(path)
        else:
            self.repo.index.add_all()
        self.repo.index.write()

    def commit(self, message: str) -> None:
        """Create a commit."""
        try:
            head = self.repo.head
            refname = head.name
            parents = [head.target]
        except pygit2.GitError:
            branch = get_default_branch()
            refname = f"refs/heads/{branch}"
            parents = []

        tree = self.repo.index.write_tree()
        author = committer = self.repo.default_signature

        self.repo.create_commit(refname, author, committer, message, tree, parents)

    def cherrypick(self, *refs: str) -> None:
        """Cherry-pick the given commits."""
        self.git("cherry-pick", *refs)

    @contextmanager
    def worktree(
        self,
        branch: str,
        path: Path,
        *,
        base: str = "HEAD",
        force: bool = False,
        force_remove: bool = False,
    ) -> Iterator[Repository]:
        """Context manager to add and remove a worktree."""
        repository = self.add_worktree(branch, path, base=base, force=force)
        try:
            yield repository
        finally:
            self.remove_worktree(path, force=force_remove)

    def add_worktree(
        self,
        branch: str,
        path: Path,
        *,
        base: str = "HEAD",
        force: bool = False,
    ) -> Repository:
        """Add a worktree."""
        self.git(
            "worktree",
            "add",
            str(path),
            "--no-track",
            "-B" if force else "-b",
            branch,
            base,
        )

        return Repository(path)

    def remove_worktree(self, path: Path, *, force: bool = False) -> None:
        """Remove a worktree."""
        if force:
            self.git("worktree", "remove", "--force", str(path))
        else:
            self.git("worktree", "remove", str(path))



"""Compatibility shims for shlex."""
import sys

if sys.version_info >= (3, 8):
    from shlex import join
else:
    from shlex import quote
    from typing import Iterable

    def join(split_command: Iterable[str]) -> str:
        """Return a shell-escaped string from *split_command*."""
        return " ".join(quote(arg) for arg in split_command)


__all__ = ["join"]



"""Command-line interface."""
import contextlib
import subprocess  # noqa: S404
import webbrowser
from pathlib import Path
from typing import Collection
from typing import Iterator
from typing import List
from typing import NoReturn
from typing import Optional

import appdirs
import click

from retrocookie.compat.contextlib import contextmanager
from retrocookie.pr import appname
from retrocookie.pr import console
from retrocookie.pr import events
from retrocookie.pr.adapters import github
from retrocookie.pr.adapters.retrocookie import retrocookie
from retrocookie.pr.base.bus import Bus
from retrocookie.pr.base.bus import Error
from retrocookie.pr.base.exceptionhandlers import ExceptionHandler
from retrocookie.pr.base.exceptionhandlers import exceptionhandler
from retrocookie.pr.base.exceptionhandlers import nullhandler
from retrocookie.pr.cache import Cache
from retrocookie.pr.core import import_pull_requests


def get_token(cache: Cache) -> str:
    """Obtain the token from the cache or by prompting the user."""
    with contextlib.suppress(FileNotFoundError):
        return cache.load_token()

    token: str = click.prompt("Please enter a personal access token for the GitHub API")
    cache.save_token(token)
    return token


def giterrorhandler(*, bus: Bus) -> ExceptionHandler:
    """Handle errors from a git subprocess."""

    @exceptionhandler
    def _(error: subprocess.CalledProcessError) -> None:
        if isinstance(error.cmd, list) and len(error.cmd) >= 2:
            command = Path(error.cmd[0]).name
            subcommand, *args = error.cmd[1:]
            if command == "git":
                bus.events.raise_(
                    events.GitFailed(
                        subcommand, args, error.returncode, error.stdout, error.stderr
                    )
                )
        return None

    return _


@exceptionhandler
def exithandler(exception: Error) -> NoReturn:
    """Exit on application errors."""
    raise SystemExit(1)


def collect(errors: List[Error]) -> ExceptionHandler:
    """Collect errors."""

    @exceptionhandler
    def _(error: Error) -> bool:
        errors.append(error)
        return True

    return _


def register_pull_request_viewer(*, bus: Bus) -> None:
    """Register an event handler for viewing pull requests in a browser."""

    @bus.events.subscribe
    def _(event: events.PullRequestCreated) -> None:
        webbrowser.open(event.template_pull.html_url)

    @bus.contexts.subscribe  # type: ignore[no-redef]
    @contextmanager
    def _(event: events.UpdatePullRequest) -> Iterator[None]:
        yield
        webbrowser.open(event.template_pull.html_url)


@click.command()
@click.argument("pull_requests", metavar="pull-request", nargs=-1)
@click.option(
    "-R",
    "--repository",
    help="GitHub repository containing the pull requests.",
)
@click.option(
    "--token",
    metavar="token",
    envvar="GITHUB_TOKEN",
    help="GitHub token",
)
@click.option("-u", "--user", help="Import pull requests opened by this GitHub user.")
@click.option("--all", is_flag=True, help="Import all pull requests.")
@click.option("--force", is_flag=True, help="Overwrite existing pull requests.")
@click.option(
    "--keep-going",
    "-k",
    is_flag=True,
    help="Keep going when some pull requests cannot be imported.",
)
@click.option(
    "--open", is_flag=True, help="View imported pull requests in a web browser."
)
@click.option("--debug", is_flag=True, help="Print tracebacks for all errors.")
@click.version_option()
def main(
    pull_requests: Collection[str],
    repository: Optional[str],
    token: Optional[str],
    user: Optional[str],
    all: bool,
    force: bool,
    keep_going: bool,
    open: bool,
    debug: bool,
) -> None:
    """Import pull requests from generated projects into their template.

    Command-line arguments specify pull requests to import, by number or by
    branch. Alternatively, select pull requests by specifying the user that
    opened them, via the --user option. Use the --all option to import all open
    pull requests in the generated project.

    Use the --repository option to specify the GitHub repository of the
    generated project from which the pull requests should be imported. Provide
    the full name of the repository on GitHub in the form owner/name. The owner
    can be omitted if the repository is owned by the authenticated user. This
    option can be omitted when the command is invoked from a local clone.

    Update previously imported pull requests by specifying --force. By default,
    this program refuses to overwrite existing pull requests.

    The program needs a personal access token (PAT) to access the GitHub API,
    and prompts for the token when invoked for the first time. Subsequent
    invocations read the token from the application cache. Alternatively, you
    can specify the token using the --token option or the GITHUB_TOKEN
    environment variable; both of these methods bypass the cache.
    """
    if not (pull_requests or user or all):
        raise click.UsageError("Please specify which pull requests to import.")

    if all and pull_requests:
        raise click.UsageError("Do not specify --all with individual pull requests.")

    bus = Bus()
    console.start(bus=bus)

    if open:  # pragma: no cover
        register_pull_request_viewer(bus=bus)

    errors: List[Error] = []

    corehandler = (
        giterrorhandler(bus=bus)
        >> github.errorhandler(bus=bus)
        >> bus.events.errorhandler()
    )
    mainerrorhandler = corehandler >> (exithandler if not debug else nullhandler)
    nestederrorhandler = (corehandler >> collect(errors)) if keep_going else nullhandler

    with mainerrorhandler:
        user_cache_dir = appdirs.user_cache_dir(appauthor=appname, appname=appname)
        cache = Cache(Path(user_cache_dir))
        api = github.API.login(token or get_token(cache), bus=bus)

        import_pull_requests(
            pull_requests,
            api=api,
            bus=bus,
            cache=cache,
            errorhandler=nestederrorhandler,
            retrocookie=retrocookie,
            repository=repository,
            user=user,
            force=force,
        )

    if errors:
        raise SystemExit(1)  # pragma: no cover


if __name__ == "__main__":  # pragma: no cover
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['retrocookie',
 'retrocookie.compat',
 'retrocookie.pr',
 'retrocookie.pr.adapters',
 'retrocookie.pr.base',
 'retrocookie.pr.compat',
 'retrocookie.pr.protocols']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.3,<9.0.0',
 'git-filter-repo>=2.26.0,<3.0.0',
 'pygit2>=1.2.1,<2.0.0']

extras_require = \
{'pr': ['typing-extensions>=3.7.4,<5.0.0',
        'github3.py>=3.0.0,<4.0.0',
        'tenacity>=6.3.1,<9.0.0',
        'appdirs>=1.4.4,<2.0.0',
        'rich>=9.5.1,<11.0.0']}

entry_points = \
{'console_scripts': ['retrocookie = retrocookie.__main__:main',
                     'retrocookie-pr = retrocookie.pr.__main__:main']}

setup_kwargs = {
    'name': 'retrocookie',
    'version': '0.4.1',
    'description': 'Update Cookiecutter templates with changes from their instances',
    'long_description': 'Retrocookie\n===========\n\n|PyPI| |Python Version| |License| |Read the Docs| |Tests| |Codecov|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/retrocookie.svg\n   :target: https://pypi.org/project/retrocookie/\n   :alt: PyPI\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/retrocookie\n   :target: https://pypi.org/project/retrocookie\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/pypi/l/retrocookie\n   :target: https://opensource.org/licenses/MIT\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/retrocookie/latest.svg?label=Read%20the%20Docs\n   :target: https://retrocookie.readthedocs.io/\n   :alt: Read the documentation at https://retrocookie.readthedocs.io/\n.. |Tests| image:: https://github.com/cjolowicz/retrocookie/workflows/Tests/badge.svg\n   :target: https://github.com/cjolowicz/retrocookie/actions?workflow=Tests\n   :alt: Tests\n.. |Codecov| image:: https://codecov.io/gh/cjolowicz/retrocookie/branch/master/graph/badge.svg\n   :target: https://codecov.io/gh/cjolowicz/retrocookie\n   :alt: Codecov\n\n\nRetrocookie updates Cookiecutter_ templates with changes from generated projects.\n\nWhen developing Cookiecutter templates,\nyou often need to work in a generated project rather than the template itself.\nReasons for this include the following:\n\n- You need to run the Continuous Integration suite for the generated project\n- Your development tools choke when running on the templated project\n\nAny changes you make in the generated project\nneed to be backported into the template,\ncarefully replacing expanded variables from ``cookiecutter.json`` by templating tags,\nand escaping any use of ``{{`` and ``}}``\nor other tokens with special meaning in Jinja.\n\nRetrocookie helps you in this situation.\n\nIt is designed to fetch commits from the repository of a generated project,\nand import them into your Cookiecutter repository,\nrewriting them on the fly to insert templating tags,\nescape Jinja-special constructs,\nand place files in the template directory.\n\nUnder the hood,\nRetrocookie rewrites the selected commits using git-filter-repo_,\nsaving them to a temporary repository.\nIt then fetches and cherry-picks the rewritten commits\nfrom the temporary repository into the Cookiecutter template,\nusing pygit2_.\n\nMaybe you\'re thinking,\nhow can this possibly work?\nOne cannot reconstruct a Jinja template from its rendered output.\nHowever, simple replacements of template variables work well in practice\nwhen you\'re only importing a handful of commits at a time.\n\n**Important:**\n\nRetrocookie relies on a ``.cookiecutter.json`` file in the generated project\nto work out how to rewrite commits.\nThis file is similar to the ``cookiecutter.json`` file in the template,\nbut contains the specific values chosen during project generation.\nYou can generate this file by putting it into the template directory in the Cookiecutter,\nwith the following contents:\n\n.. code:: jinja\n\n   {{ cookiecutter | jsonify }}\n\n\nRequirements\n------------\n\n* Python 3.7+\n* git >= 2.22.0\n\n\nInstallation\n------------\n\nYou can install *Retrocookie* via pip_ from PyPI_:\n\n.. code:: console\n\n   $ pip install retrocookie\n\nOptionally, install the ``pr`` extra for the retrocookie-pr_ command:\n\n.. code:: console\n\n   $ pip install retrocookie[pr]\n\n\nExample\n-------\n\nHere\'s an example to demonstrate the general workflow.\n\nTo start with, we clone the repository of your Cookiecutter template.\nFor this example, I\'ll use my own `Hypermodern Python Cookiecutter`_.\n\n.. code:: console\n\n   $ git clone https://github.com/cjolowicz/cookiecutter-hypermodern-python\n\nNext, we\'ll create a project from the template,\nand set up a git repository for it:\n\n.. code:: console\n\n   $ cookiecutter --no-input cookiecutter-hypermodern-python\n   $ cd hypermodern-python\n   $ git init\n   $ git add .\n   $ git commit --message="Initial commit"\n\nLet\'s open a feature branch in the project repository,\nand make a fictitious change involving the default project name *hypermodern-python*:\n\n.. code:: console\n\n   $ git switch --create add-example\n   $ echo \'# hypermodern-python\' > EXAMPLE.md\n   $ git add EXAMPLE.md\n   $ git commit --message="Add example"\n\nBack in the Cookiecutter repository,\nwe can now invoke retrocookie to import the changes from the feature branch:\n\n.. code:: console\n\n   $ cd ../cookiecutter-hypermodern-python\n   $ retrocookie --branch add-example --create ../hypermodern-python\n\nA ``git show`` in the Cookiecutter shows the file under the template directory,\non a branch named as in the original repository,\nwith the project name replaced by a Jinja tag:\n\n.. code:: diff\n\n   commit abb4f823b9f1760e3a678c927ec9797c0a40a9b6 (HEAD -> add-example)\n   Author: Your Name <your.name@example.com>\n   Date:   Fri Dec 4 23:40:41 2020 +0100\n\n       Add example\n\n   diff --git a/{{cookiecutter.project_name}}/EXAMPLE.md b/{{cookiecutter.project_name}}/EXAMPLE.md\n   new file mode 100644\n   index 0000000..a158618\n   --- /dev/null\n   +++ b/{{cookiecutter.project_name}}/EXAMPLE.md\n   @@ -0,0 +1 @@\n   +# {{cookiecutter.project_name}}\n\n\nUsage\n-----\n\nThe basic form:\n\n.. code::\n\n   $ retrocookie <repository> [<commits>...]\n   $ retrocookie <repository> -b <branch> [--create]\n\nThe ``<repository>`` is a filesystem path to the source repository.\nFor ``<commits>``, see `gitrevisions(7)`__.\n\n__ https://git-scm.com/docs/gitrevisions\n\nImport ``HEAD`` from ``<repository>``:\n\n.. code::\n\n   $ retrocookie <repository>\n\nImport the last two commits:\n\n.. code::\n\n   $ retrocookie <repository> HEAD~2..\n\nImport by commit hash:\n\n.. code::\n\n   $ retrocookie <repository> 53268f7 6a3368a c0b4c6c\n\nImport commits from branch ``topic``:\n\n.. code::\n\n   $ retrocookie <repository> --branch=topic\n\nEquivalently:\n\n.. code::\n\n   $ retrocookie <repository> master..topic\n\nImport commits from ``topic`` into a branch with the same name:\n\n.. code::\n\n   $ retrocookie <repository> --branch=topic --create\n\nEquivalently, using short options:\n\n.. code::\n\n   $ retrocookie <repository> -cb topic\n\nImport commits from branch ``topic``, which was branched off ``1.0``:\n\n.. code::\n\n   $ retrocookie <repository> --branch=topic --upstream=1.0\n\nEquivalently:\n\n.. code::\n\n   $ retrocookie <repository> 1.0..topic\n\nImport ``HEAD`` into a new branch ``topic``:\n\n.. code::\n\n   $ retrocookie <repository> --create-branch=topic\n\nPlease see the `Command-line Reference <Usage_>`_ for further details.\n\n\n.. _retrocookie-pr:\n\nImporting pull requests from generated projects with retrocookie-pr\n-------------------------------------------------------------------\n\nYou can import pull requests from a generated project to the project template,\nassuming their repositories are on GitHub_.\nThis requires activating the ``pr`` extra when installing with pip_:\n\n.. code::\n\n  $ pip install retrocookie[pr]\n\nThe command ``retrocookie-pr`` has the basic form:\n\n.. code::\n\n   $ retrocookie-pr [-R <repository>] [<pr>...]\n   $ retrocookie-pr [-R <repository>] --user=<user>\n   $ retrocookie-pr [-R <repository>] --all\n\nCommand-line arguments specify pull requests to import, by number or by branch.\nPull requests from forks are currently not supported.\n\nUse the ``-R <repository>`` option to specify the GitHub repository of the generated project\nfrom which the pull requests should be imported.\nProvide the full name of the repository on GitHub in the form ``owner/name``.\nThe owner can be omitted if the repository is owned by the authenticated user.\nThis option can be omitted when the command is invoked from a local clone.\n\nYou can also select pull requests by specifying the user that opened them, via the ``--user`` option.\nThis is handy for importing automated pull requests, such as dependency updates from Dependabot_.\n\nUse the ``--all`` option to import all open pull requests in the generated project.\n\nYou can update previously imported pull requests by specifying ``--force``.\nBy default, ``retrocookie-pr`` refuses to overwrite existing pull requests.\n\nThe command needs a `personal access token`_ to access the GitHub API.\n(This token is also used to push to the GitHub repository of the project template.)\nYou will be prompted for the token when you invoke the command for the first time.\nOn subsequent invocations, the token is read from the application cache.\nAlternatively, you can specify the token using the ``--token`` option or the ``GITHUB_TOKEN`` environment variable;\nboth of these methods bypass the cache.\n\nUse the ``--open`` option to open each imported pull request in a web browser.\n\nPlease see the `Command-line Reference <Usage_>`_ for further details.\n\n\nContributing\n------------\n\nContributions are very welcome.\nTo learn more, see the `Contributor Guide`_.\n\n\nLicense\n-------\n\nDistributed under the terms of the MIT_ license,\n*Retrocookie* is free and open source software.\n\n\nIssues\n------\n\nIf you encounter any problems,\nplease `file an issue`_ along with a detailed description.\n\n\nCredits\n-------\n\nThis project was generated from `@cjolowicz`_\'s `Hypermodern Python Cookiecutter`_ template.\n\n\n.. _@cjolowicz: https://github.com/cjolowicz\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _Dependabot: https://dependabot.com/\n.. _GitHub: https://github.com/\n.. _Hypermodern Python Cookiecutter: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n.. _MIT: http://opensource.org/licenses/MIT\n.. _PyPI: https://pypi.org/\n.. _file an issue: https://github.com/cjolowicz/retrocookie/issues\n.. _git-filter-repo: https://github.com/newren/git-filter-repo\n.. _git rebase: https://git-scm.com/docs/git-rebase\n.. _pip: https://pip.pypa.io/\n.. _personal access token: https://docs.github.com/en/free-pro-team@latest/github/authenticating-to-github/creating-a-personal-access-token\n.. _pygit2: https://github.com/libgit2/pygit2\n.. github-only\n.. _Contributor Guide: CONTRIBUTING.rst\n.. _Usage: https://retrocookie.readthedocs.io/en/latest/usage.html\n',
    'author': 'Claudio Jolowicz',
    'author_email': 'mail@claudiojolowicz.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cjolowicz/retrocookie',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Events and contexts for the message bus."""
from dataclasses import dataclass
from typing import List

from retrocookie import git
from retrocookie.pr.base import bus
from retrocookie.pr.protocols import github


@dataclass
class GitNotFound(bus.Event):
    """Cannot find an installation of git."""


@dataclass
class BadGitVersion(bus.Event):
    """The installed version of git is too old."""

    version: git.Version
    expected: git.Version


@dataclass
class ProjectNotFound(bus.Event):
    """The generated project could not be identified."""


@dataclass
class TemplateNotFound(bus.Event):
    """The project template could not be identified."""

    project: github.Repository


@dataclass
class LoadProject(bus.Context):
    """The project repository is being loaded."""

    repository: str


@dataclass
class LoadTemplate(bus.Context):
    """A template repository is being loaded."""

    repository: str


@dataclass
class RepositoryNotFound(bus.Event):
    """The repository was not found."""

    repository: str


@dataclass
class PullRequestNotFound(bus.Event):
    """The pull request was not found."""

    pull_request: str


@dataclass
class PullRequestAlreadyExists(bus.Event):
    """The pull request already exists."""

    template: github.Repository
    template_pull: github.PullRequest
    project_pull: github.PullRequest


@dataclass
class GitFailed(bus.Event):
    """The git command exited with a non-zero status."""

    command: str
    options: List[str]
    status: int
    stdout: str
    stderr: str


@dataclass
class GitHubError(bus.Event):
    """The GitHub API returned an error response."""

    url: str
    method: str
    code: int
    message: str
    errors: List[str]


@dataclass
class ConnectionError(bus.Event):
    """A connection to the GitHub API could not be established."""

    url: str
    method: str
    error: str


@dataclass
class CreatePullRequest(bus.Context):
    """A pull request is being created."""

    template: github.Repository
    template_branch: str
    project_pull: github.PullRequest


@dataclass
class UpdatePullRequest(bus.Context):
    """A pull request is being updated."""

    template: github.Repository
    template_pull: github.PullRequest
    project_pull: github.PullRequest


@dataclass
class PullRequestCreated(bus.Event):
    """A pull request was created."""

    template: github.Repository
    template_pull: github.PullRequest
    project_pull: github.PullRequest



"""Interface for git-filter-repo."""
import contextlib
import io
import re
from pathlib import Path
from typing import Any
from typing import Container
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import overload
from typing import Sequence
from typing import Tuple

from git_filter_repo import Blob
from git_filter_repo import FilteringOptions
from git_filter_repo import RepoFilter

from . import git


def get_replacements(
    context: Dict[str, Any],
    include_variables: Container[str],
    exclude_variables: Container[str],
) -> List[Tuple[bytes, bytes]]:
    """Return replacements to be applied to commits from the template instance."""

    def ref(key: str) -> str:
        return f"{{{{cookiecutter.{key}}}}}"

    return [
        (value.encode(), ref(key).encode())
        for key, value in context.items()
        if key not in exclude_variables
        and not (include_variables and key not in include_variables)
        and isinstance(value, str)
    ]


def find_token(
    string: bytes, pos: int, tokens: Sequence[bytes]
) -> Tuple[Optional[bytes], int]:
    """Find the first occurrence of any of multiple tokens."""
    pattern = re.compile(b"|".join(re.escape(token) for token in tokens))
    match = pattern.search(string, pos)
    if match is None:
        return None, -1
    return match.group(), match.start()


def quote_tokens(
    text: bytes, quotes: Tuple[bytes, bytes], tokens: Sequence[bytes]
) -> bytes:
    """Wrap tokens in ``<quotes[0]><token><quotes[1]>``."""

    def _generate() -> Iterator[bytes]:
        index = 0

        while index < len(text):
            token, found = find_token(text, index, tokens)

            if token is None:
                yield text[index:]
                break

            if found != 0:
                yield text[index:found]

            yield token.join(quotes)
            index = found + len(token)

    return b"".join(_generate())


@overload
def to_bytes(args: Tuple[str, str]) -> Tuple[bytes, bytes]:  # noqa: D103
    ...  # pragma: no cover


@overload
def to_bytes(args: Tuple[str, ...]) -> Tuple[bytes, ...]:  # noqa: D103
    ...  # pragma: no cover


def to_bytes(args: Tuple[str, ...]) -> Tuple[bytes, ...]:
    """Convert each str to bytes."""
    return tuple(arg.encode() for arg in args)


def escape_jinja(text: bytes) -> bytes:
    """Escape Jinja tokens."""
    quotes = to_bytes(('{{"', '"}}'))
    tokens = to_bytes(("{{", "}}", "{%", "%}", "{#", "#}"))
    return quote_tokens(text, quotes, tokens)


class RepositoryFilter:
    """Perform path and blob replacements on a repository."""

    def __init__(
        self,
        repository: git.Repository,
        source: git.Repository,
        commits: Iterable[str],
        template_directory: Path,
        context: Dict[str, Any],
        include_variables: Container[str],
        exclude_variables: Container[str],
    ) -> None:
        """Initialize."""
        self.repository = repository
        self.source = source
        self.commits = commits
        self.template_directory = str(template_directory).encode()
        self.replacements = get_replacements(
            context, include_variables, exclude_variables
        )

    def filename_callback(self, filename: bytes) -> bytes:
        """Rewrite filenames."""
        for old, new in self.replacements:
            filename = filename.replace(old, new)
        return b"/".join((self.template_directory, filename))

    def blob_callback(self, blob: Blob, metadata: Dict[str, Any]) -> None:
        """Rewrite blobs."""
        blob.data = escape_jinja(blob.data)
        for old, new in self.replacements:
            blob.data = blob.data.replace(old, new)

    def _create_filter(self) -> RepoFilter:
        """Create the filter."""
        args = FilteringOptions.parse_args(
            [
                f"--source={self.source.path}",
                f"--target={self.repository.path}",
                "--replace-refs=update-and-add",
                "--refs",
                *self.commits,
            ]
        )
        return RepoFilter(
            args,
            filename_callback=self.filename_callback,
            blob_callback=self.blob_callback,
        )

    def run(self) -> None:
        """Run the filter."""
        null = io.StringIO()
        with contextlib.redirect_stdout(null), contextlib.redirect_stderr(null):
            repofilter = self._create_filter()
            repofilter.run()



"""Retrocookie interface."""
from pathlib import Path

from retrocookie import retrocookie as _retrocookie


def retrocookie(repository: Path, *, branch: str, upstream: str, path: Path) -> None:
    """Wrap retrocookie.retrocookie with protocols.Retrocookie signature."""
    _retrocookie(repository, branch=branch, upstream=upstream, path=path)



"""Command-line interface."""
from pathlib import Path
from typing import Container
from typing import Iterable
from typing import Optional

import click

from .core import retrocookie


@click.command()
@click.option(
    "--branch",
    "-b",
    metavar="BRANCH",
    help="Remote branch to cherry-pick",
)
@click.option(
    "--upstream",
    metavar="REF",
    help="Remote upstream branch",
)
@click.option(
    "--create",
    "-c",
    is_flag=True,
    help="Create a local branch with the same name as --branch",
)
@click.option(
    "--create-branch",
    metavar="BRANCH",
    help="Create a local branch for the imported commits",
)
@click.option(
    "--include-variable",
    "include_variables",
    metavar="VAR",
    multiple=True,
    help="Only rewrite these Cookiecutter variables",
)
@click.option(
    "--exclude-variable",
    "exclude_variables",
    metavar="VAR",
    multiple=True,
    help="Do not rewrite these Cookiecutter variables",
)
@click.option(
    "--directory",
    "-C",
    metavar="DIR",
    help="Path to the repository  [default: working directory]",
)
@click.argument("repository")
@click.argument("commits", nargs=-1)
@click.version_option()
def main(
    branch: Optional[str],
    upstream: Optional[str],
    create: bool,
    create_branch: Optional[str],
    include_variables: Container[str],
    exclude_variables: Container[str],
    directory: Optional[str],
    repository: str,
    commits: Iterable[str],
) -> None:
    """Retrocookie imports commits into Cookiecutter templates.

    The path to the source repository is passed as a positional argument.
    This should be an instance of the Cookiecutter template with a
    .cookiecutter.json file.

    Additional positional arguments are the commits to be imported.
    See gitrevisions(7) for a list of ways to spell commits and commit
    ranges. These arguments are interpreted in the same way as
    git-cherry-pick(1) does. If no commits are specified, the HEAD
    commit is imported.

    Specifying --branch is equivalent to passing master..<branch>
    as a positional argument. Use --upstream to specify a different
    upstream branch. Use --create to create the same branch in the
    target repository.

    By default, commits are cherry-picked onto the current branch.
    Use --create-branch to specify a different branch, which must
    not exist yet. As a shorthand, --create is equivalent to passing
    --create-branch with the same argument as --branch.

    Commits are rewritten in the following way:

    \b
      - Files are moved to the template directory
      - Tokens with special meaning in Jinja are escaped
      - Values from .cookiecutter.json are replaced by templating tags
        with the corresponding variable name

    Use the --include-variable and --exclude-variable options to include or
    exclude specific variables from .cookiecutter.json.
    """
    if create:
        if create_branch:
            raise click.UsageError(
                "--create and --create-branch are mutually exclusive"
            )

        if branch is None:
            raise click.UsageError("--create requires --branch")

        create_branch = branch

    if not commits and branch is None:
        commits = ["HEAD"]

    path = Path(directory) if directory else None
    retrocookie(
        Path(repository),
        commits,
        branch=branch,
        upstream=upstream,
        create_branch=create_branch,
        include_variables=include_variables,
        exclude_variables=exclude_variables,
        path=path,
    )


if __name__ == "__main__":
    main(prog_name="retrocookie")  # pragma: no cover



"""Protocols for retrocookie.pr.github."""
from __future__ import annotations

from typing import AbstractSet
from typing import Iterable
from typing import Optional
from typing import Set

from retrocookie.pr.compat.typing import Protocol


class PullRequest(Protocol):
    """Pull request."""

    @property
    def number(self) -> int:
        """The number of the pull request."""

    @property
    def title(self) -> str:
        """The title of the pull request."""

    @property
    def body(self) -> Optional[str]:
        """The body of the pull request."""

    @property
    def branch(self) -> str:
        """The branch merged by the pull request."""

    @property
    def user(self) -> str:
        """The user who opened the pull request."""

    @property
    def html_url(self) -> str:
        """URL for viewing the pull request in a browser."""

    @property
    def labels(self) -> Set[str]:
        """Labels applied to the pull request."""

    def update(self, title: str, body: Optional[str], labels: AbstractSet[str]) -> None:
        """Update the pull request."""


class Repository(Protocol):
    """GitHub repository."""

    @property
    def owner(self) -> str:
        """The user who owns the repository."""

    @property
    def full_name(self) -> str:
        """The full name of the repository."""

    @property
    def clone_url(self) -> str:
        """URL for cloning the repository."""

    @property
    def push_url(self) -> str:
        """URL for pushing to the repository."""

    @property
    def default_branch(self) -> str:
        """The default branch of the repository."""

    def pull_requests(self) -> Iterable[PullRequest]:
        """Pull requests open in the repository."""

    def pull_request(self, number: int) -> PullRequest:
        """Return pull request identified by the given number."""

    def pull_request_by_head(self, head: str) -> Optional[PullRequest]:
        """Return pull request for the given head."""

    def create_pull_request(
        self,
        *,
        head: str,
        title: str,
        body: Optional[str],
        labels: AbstractSet[str],
    ) -> PullRequest:
        """Create a pull request."""


class API(Protocol):
    """GitHub API."""

    @property
    def me(self) -> str:
        """Return the login of the authenticated user."""

    def repository(self, owner: str, name: str) -> Repository:
        """Return the repository with the given owner and name."""
